﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Linq;
using System.ComponentModel;
using System.Threading;

using AllButOneProduct;

namespace AllButOneProductView
{
	public class Graph : INotifyPropertyChanged
	{
		private string _Name;
		private Tuple<int, long>[] _Values;

		public string Name
		{
			get { return _Name; }
			set { _Name = value; FirePropertyChanged("Name"); }
		}

		public Tuple<int, long>[] Values
		{
			get { return _Values; }
			set { _Values = value; FirePropertyChanged("Values"); }
		}

		#region INotifyPropertyChanged Members

		public event PropertyChangedEventHandler PropertyChanged;

		void FirePropertyChanged(string propertyName)
		{
			PropertyChangedEventHandler d = PropertyChanged;
			if (d == null)
			{
				return;
			}
			d(this, new PropertyChangedEventArgs(propertyName));
		}

		#endregion
	}

	public class ViewModel : INotifyPropertyChanged
	{
		const int MAX = 40;
		const int REPEATS = 500000;
		const int SLOW_CUT = 20;
		const int NUM_GRAPHS = 4;

		private int stillRunning = NUM_GRAPHS;

		public ViewModel()
		{
			Graph[] graphs = new Graph[4] {
				new Graph { Name = "N Squared", Values = ZeroSeries() },
				new Graph { Name = "With Division", Values = ZeroSeries() },
				new Graph { Name = "Left Right Products", Values = ZeroSeries() },
				new Graph { Name = "Fast Right Left Scan", Values = ZeroSeries() }
			};

			SynchronizationContext sc = SynchronizationContext.Current;

			ThreadPool.QueueUserWorkItem((o) =>
			{
				var r1 = ConvertToChartSeries(AllButOneProductTiming.CalcProductsWithDivisionTimes(MAX, REPEATS));
				sc.Post((o2) => { graphs[1].Values = r1; OneComplete(); }, null);

				var r2 = ConvertToChartSeries(AllButOneProductTiming.CalcProductsFromLeftRightProductsTimes(MAX, REPEATS));
				sc.Post((o2) => { graphs[2].Values = r2; OneComplete(); }, null);

				var r3 = ConvertToChartSeries(AllButOneProductTiming.CalcProductsWithFastRightLeftScanTimes(MAX, REPEATS));
				sc.Post((o2) => { graphs[3].Values = r3; OneComplete(); }, null);

				var r0 = ConvertToChartSeries(AllButOneProductTiming.CalcProductsNSquaredTimes(MAX, REPEATS / SLOW_CUT), SLOW_CUT);
				sc.Post((o2) => { graphs[0].Values = r0; OneComplete(); }, null);
			});

			Graphs = graphs;
		}

		private void OneComplete()
		{
			if (Interlocked.Decrement(ref stillRunning) == 0)
				FirePropertyChanged("ShowBusy");
		}

		public Graph[] Graphs { get; private set; }

		public bool ShowBusy { get { return stillRunning > 0; } }

		private static Tuple<int, long>[] ZeroSeries()
		{
			return Enumerable.Range(0, MAX).Select((i) => new Tuple<int, long>(i, 0)).ToArray();
		}

		private static Tuple<int, long>[] ConvertToChartSeries(long[] times, int multiplyUp = 1)
		{
			return times.Select((time, i) => new Tuple<int, long>(i, time * multiplyUp)).ToArray();
		}

		// from CLR via C# p772
		private static AsyncCallback SyncContextCallback(AsyncCallback callback)
		{
			// Capture the calling thread's SynchronizationContext-derived object
			SynchronizationContext sc = SynchronizationContext.Current;

			// If there is no SC, just return what was passed in
			if (sc == null)
				return callback;

			// Return aq delegate that, when invoked, posts to the captured SC a method that
			// calls the original AsyncCallback passing it the IAsyncResult argument
			return asyncResult => sc.Post(result => callback((IAsyncResult)result), asyncResult);
		}

		#region INotifyPropertyChanged Members

		public event PropertyChangedEventHandler PropertyChanged;

		void FirePropertyChanged(string propertyName)
		{
			PropertyChangedEventHandler d = PropertyChanged;
			if (d == null)
			{
				return;
			}
			d(this, new PropertyChangedEventArgs(propertyName));
		}

		#endregion
	}
}
